# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Expands datetime features from input into sub features."""
from typing import Optional, Union
from logging import Logger
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from automl.client.core.common.logging_utilities import function_debug_log_wrapped
from automl.client.core.common.featurization.logger import TransformerLogger


class DateTimeFeaturesTransformer(BaseEstimator, TransformerMixin, TransformerLogger):
    """
    Expands datetime features from input into sub features.

    Like year, month, day, day of the week, day of the year, quarter, week of
    the month, hour, minute and second.
    """

    def __init__(self, logger: Optional[Union[Logger, TransformerLogger]] = None):
        """
        Initialize the logger.

        :param logger: Logger to be injected to usage in this class.
        """
        self._init_logger(logger)

    def fit(self, x, y=None):
        """
        Fit function for date time transform.

        :param x: Input array.
        :type x: numpy.ndarray or pandas.series
        :param y: Target values.
        :type y: numpy.ndarray
        :return: The instance object: self.
        """
        return self

    @function_debug_log_wrapped
    def transform(self, x):
        """
        Transform data x.

        :param x: The data to transform.
        :type x: numpy.ndarray or pandas.series
        :return: The transformed data.
        """
        return self._datetime_feats(x)

    def _datetime_feats(self, x):
        """
        Get the features for a datetime column.

        Expand the date time features from array of dates.

        :param x: Series that represents column.
        :type x: numpy.ndarray or pandas.series
        :return: Features for datetime column.
        """
        x = pd.to_datetime(pd.Series(x), infer_datetime_format=True,
                           box=False, errors="coerce").fillna(pd.Timestamp.min)
        return pd.concat([
            x.dt.year,
            x.dt.month,
            x.dt.day,
            x.dt.dayofweek,
            x.dt.dayofyear,
            x.dt.quarter,
            x.apply(lambda dt: (dt.day - 1) // 7 + 1),
            x.dt.hour,
            x.dt.minute,
            x.dt.second,
        ], axis=1).values
